package redactableBlockchain;
import java.math.BigInteger;
import java.security.*;
import java.security.spec.ECGenParameterSpec;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.bouncycastle.jcajce.provider.asymmetric.ec.BCECPublicKey;
import org.bouncycastle.jce.ECNamedCurveTable;
import org.bouncycastle.jce.spec.ECParameterSpec;
import org.bouncycastle.jce.spec.ECPrivateKeySpec;
import org.bouncycastle.jce.spec.ECPublicKeySpec;
import org.bouncycastle.math.ec.ECPoint;

import com.google.gson.Gson;

//import myChain.old.NoobChain;
import redactableBlockchain.data.transaction.PaymentTransaction;
import redactableBlockchain.data.transaction.TransactionInput;
import redactableBlockchain.data.transaction.TransactionOutput;
import redactableBlockchain.util.StringUtil;
import redactableBlockchain.util.Utils;

public class Wallet {

	public PrivateKey privateKey;//TODO
	private PublicKey publicKey;

	public PublicKey getPublicKey() {
		return publicKey;
	}

	public String getPublicKeyAsString() throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		return Utils.PublicKeyToBase64(publicKey);
	}

	private HashMap<String,TransactionOutput> UTXOs = new HashMap<String,TransactionOutput>();

	public Wallet() {
		generateKeyPair();
	}

	public Wallet(String seed) {
		generateKeyPairFromSeed(seed);
	}

	private void generateKeyPairFromSeed(String seed) {
		try {
			KeyPairGenerator keyGen = KeyPairGenerator.getInstance("ECDSA","BC");
			SecureRandom random = SecureRandom.getInstance("SHA1PRNG");
			if(seed != null)
				random.setSeed(seed.getBytes("UTF-8"));
			ECGenParameterSpec ecSpec = new ECGenParameterSpec("prime192v1");
			// Initialize the key generator and generate a KeyPair
			keyGen.initialize(ecSpec, random); //256 
			KeyPair keyPair = keyGen.generateKeyPair();
			// Set the public and private keys from the keyPair
			privateKey = keyPair.getPrivate();
			publicKey = keyPair.getPublic();

		}catch(Exception e) {
			throw new RuntimeException(e);
		}
	}

	public void generateKeyPair() {
		generateKeyPairFromSeed(null);

	}

	public float getBalance() {
		return (Float) null;
		//		float total = 0;	
		//		HashMap<String, TransactionOutput> allUTXOs = BlockChainImpl.getMyUTXOs(publicKey);
		//        for (TransactionOutput UTXO: allUTXOs.values())
		//        {        	
		//            try {
		//				if(UTXO.isMine(publicKey)) { 
		//					UTXOs.put(UTXO.id,UTXO);
		//					total += UTXO.value ; 
		//				}
		//			} catch (Exception e) {
		//				e.printStackTrace();
		//			}
		//        }  
		//		return total;
	}

//	public PaymentTransaction sendFunds(PublicKey _recipient, int value ) throws NoSuchAlgorithmException, InvalidKeySpecException {
//		if(getBalance() < value) {
//			System.out.println("#Not Enough funds to send transaction. Transaction Discarded.");
//			return null;
//		}
//		ArrayList<TransactionInput> inputs = new ArrayList<TransactionInput>();
//
//		float total = 0;
//		for (Map.Entry<String, TransactionOutput> item: UTXOs.entrySet()){
//			TransactionOutput UTXO = item.getValue();
//			total += UTXO.value;
//			inputs.add(new TransactionInput(UTXO.id));
//			if(total > value) break;
//		}
//
//		PaymentTransaction newTransaction = new PaymentTransaction(publicKey, _recipient , value, inputs);
//		newTransaction.sign(privateKey);
//
//		for(TransactionInput input: inputs){
//			UTXOs.remove(input.transactionOutputId);
//		}
//
//		return newTransaction;
//	}

	public String getJson()
	{
		Gson gson = new Gson();
		String keyPairJson = gson.toJson(privateKey);
		return keyPairJson;
	}

	public byte[] sign(String data) {
		return StringUtil.applyECDSASig(privateKey,data);
	}
}


