package redactableBlockchain.data;

import java.math.BigInteger;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;

import redactableBlockchain.data.transaction.Transaction;
import redactableBlockchain.exception.BlockInvalidException;

public abstract class Block implements IBlock {

	protected long nonce;
	protected String merkleRoot;
	protected String hash;
	protected ArrayList<Transaction> transactions = new ArrayList<Transaction>(); //our data will be a simple message.
	public long timeStamp; //as number of milliseconds since 1/1/1970.
	private BlockType blockType;
	
	public Block(BlockType type)
	{
		this.blockType = type;
	}
	
	@Override
	public final ArrayList<Transaction> getTransactions() {
		return transactions;
	}

	@Override
	public long getCreationTime() {
		return timeStamp;
	}

	@Override
	public String getMerkleRoot() {
		return this.merkleRoot;
	}

	@Override
	public void setNonce(long nonce) {
		this.nonce = nonce;
	}
	
	public BigInteger getTargetDiff() {
		return max.divide(BigInteger.valueOf(16*16*16));
	}

	public BigInteger getDiff() {
		return max.divide(new BigInteger(hash, 16));
	}

	public String getHash() {//TODO remove?
		String hash = calculateHash();
		return hash;
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((blockType == null) ? 0 : blockType.hashCode());
		result = prime * result + ((hash == null) ? 0 : hash.hashCode());
		return result;
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Block other = (Block) obj;
		if (blockType != other.blockType)
			return false;
		if (hash == null) {
			if (other.hash != null)
				return false;
		} else if (!hash.equals(other.hash))
			return false;
		return true;
	}


	
}
