package redactableBlockchain.data;

import org.apache.commons.collections4.OrderedMapIterator;
import org.apache.commons.collections4.map.LinkedMap;

import redactableBlockchain.data.transaction.ElectionTransaction;
import redactableBlockchain.data.transaction.Transaction;
import redactableBlockchain.util.Constants;

public class Chain {

	private LinkedMap<String, IBlock> myChain;
	private int votingActive = -1;
	private ElectionTransaction activeElection = null;
	private int inFavorOfElection = 0;
	private String nextBlockMustReferenceModifyBlock = null;
	private boolean createModifyBlock;

	public Chain(Chain myChain2) {
		this.myChain = new LinkedMap<>(myChain2.myChain);
	}

	public Chain() {
		myChain = new LinkedMap<>();
	}

	public OrderedMapIterator<String, IBlock> mapIterator() {
		return myChain.mapIterator();
	}

	public IBlock get(String key) {
		return myChain.get(key);
	}

	public void setVotingActive(int i) {
		votingActive = i;
	}

	public int getVotingActive() {
		return votingActive;
	}

	public void setInFavorOfElection(int i) {
		inFavorOfElection = i;
	}

	public int getInFavorOfElection() {
		return inFavorOfElection;
	}

	public void put(String hash, IBlock block) {
		myChain.put(hash, block);
	}

	public String lastKey() {
		return myChain.lastKey();
	}
	
	public IBlock getLastBlock()
	{
		return myChain.get(myChain.lastKey());
	}

	public int size() {
		return myChain.size();
	}

	public void printChain() {
		StringBuilder sb = new StringBuilder();
		int height = 0;
		for(String key : myChain.keySet())
		{
			IBlock b = myChain.get(key);
			sb.append(height+" "+b+"\n");
			height++;
		}
		System.out.println("Chain:\n"+sb.toString());
	}

	public String get(int index) {
		return myChain.get(index);
	}

	public int indexOf(String key) {
		return myChain.indexOf(key);
	}

	public String getActiveVoting() {
		int lastIndex = myChain.indexOf(myChain.lastKey());
		//TODO
		for(int i = lastIndex;i>lastIndex-Constants.votingDuration && i>=0;i--)
		{
			String key = myChain.get(i);
			IBlock b = myChain.get(key);
			ElectionTransaction tx = blockContainsElection(b);
			if(tx != null)
			{
				return tx.calculateHash();
			}
		}
		return null;
	}

	private ElectionTransaction blockContainsElection(IBlock b) {
		for(Transaction tx : b.getTransactions())
		{
			if(tx instanceof ElectionTransaction)
			{
				return (ElectionTransaction) tx;
			}
		}
		return null;
	}

	public ElectionTransaction getEndedVoting() {
		int lastIndex = myChain.indexOf(myChain.lastKey());
		String key = myChain.get(lastIndex-Constants.votingDuration);
		IBlock b = myChain.get(key );
		ElectionTransaction tx = blockContainsElection(b);
		return tx;
	}

	public ElectionTransaction getActiveElection() {
		return activeElection;
	}

	public void setActiveElection(ElectionTransaction activeElection) {
		this.activeElection = activeElection;
	}

	public String getNextBlockMustReferenceModifyBlock() {
		return nextBlockMustReferenceModifyBlock;
	}

	public void setNextBlockMustReferenceModifyBlock(String nextBlockMustReferenceModifyBlock) {
		this.nextBlockMustReferenceModifyBlock = nextBlockMustReferenceModifyBlock;
	}

	public void setCreateModifyBlock(boolean b) {
		this.createModifyBlock = b;
	}

	public final boolean isCreateModifyBlock() {
		return createModifyBlock;
	}

	public String getNextRedactionBlockHash() {
		String lastRedactionBlockHash = null;
		if(nextBlockMustReferenceModifyBlock != null)
			return nextBlockMustReferenceModifyBlock;
		if(myChain.size() == 1)//only genesisblock
		{
			lastRedactionBlockHash = Constants.getGenesisBlock().getHash();
		}
		else
		{
			IBlock previousBlock = getLastBlock();
			lastRedactionBlockHash = ((StandardBlock)previousBlock).getLastRedactionBlockHash();
		}
		return lastRedactionBlockHash;
	}
	
}
