package redactableBlockchain.data;

import java.io.Serializable;
import java.math.BigInteger;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import redactableBlockchain.data.transaction.PaymentTransaction;
import redactableBlockchain.data.transaction.Transaction;
import redactableBlockchain.data.transaction.TransactionOutput;
import redactableBlockchain.data.transaction.VoteQuestion;
import redactableBlockchain.data.transaction.VotingTransaction;
import redactableBlockchain.exception.BlockInvalidException;
import redactableBlockchain.exception.VerificationException;
import redactableBlockchain.util.Constants;
import redactableBlockchain.util.StringUtil;

public class CorrectionBlock extends Block implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = -6803986682848737624L;
	private String previousHashChain2; // hash of the previous block in the second/redact chain
	private String lastBlockChain1Hash;// hash of the last block in the first/normal chain
	private String electionHash;

	private static final Logger logger = LoggerFactory.getLogger(CorrectionBlock.class);
	//Block Constructor.

	public CorrectionBlock(String previousBlockChain1Hash, String subsequentBlockChain1Hash,
			String lastBlockChain1Hash, String previousHashChain2, String electionHash, Transaction coinbase) {
		super(BlockType.ModifyBlock);
		this.previousHashChain2 = previousHashChain2;
//		this.previousBlockChain1Hash = previousBlockChain1Hash;
//		this.subsequentBlockChain1Hash = subsequentBlockChain1Hash;
		this.lastBlockChain1Hash = lastBlockChain1Hash;
		this.electionHash = (electionHash);//TODO include in block hash
		this.timeStamp = new Date().getTime();
		this.transactions.add(coinbase);
//		prepareBlockForMining();
		merkleRoot = StringUtil.getMerkleRoot(transactions);
		this.hash = calculateHash(); //Making sure we do this after we set the other values.
	}

	//Calculate new hash based on blocks contents
	public String calculateHash() {
		String calculatedhash = StringUtil.applySha256( 
				previousHashChain2 +
				lastBlockChain1Hash +
				Long.toString(timeStamp) +
				Long.toString(nonce) + 
				merkleRoot
				);
		this.hash = calculatedhash;
		return calculatedhash;
	}

	//Add transactions to this block
	public boolean addTransaction(Transaction transaction) throws VerificationException {
		//process transaction and check if valid, unless block is genesis block then ignore.
		if(transaction == null) return false;		
		//		if((!"0000000000000000000000000000000000000000000000000000000000000000".equals(previousHash))) {
		//			if((transaction.processTransaction() != true)) {
		//				logger.warn("Transaction failed to process. Discarded.");
		//				return false;
		//			}
		//		}

		transactions.add(transaction);
		logger.info("Transaction Successfully added to Block");
//		prepareBlockForMining();
		merkleRoot = StringUtil.getMerkleRoot(transactions);
		return true;
	}

	public void validate() throws BlockInvalidException {
		//throws Exception ???
		calculateHash();
		//todo vorgänger
		if(getDiff().compareTo(getTargetDiff()) != -1)
		{
			System.out.println(calculateHash());
			System.out.println(getDiff());
			System.out.println(getTargetDiff());
			throw new BlockInvalidException("Difficulty too low...");
		}
		//TODO check block
		if(getTransactions().size() != 1)
		{
			throw new BlockInvalidException("Not exactly one (coinbase) transaction in block");
		}
		Transaction claimedCoinBaseTX = getTransactions().iterator().next();
		if(claimedCoinBaseTX instanceof PaymentTransaction)
		{
			PaymentTransaction tx = (PaymentTransaction) claimedCoinBaseTX;
			if(tx.inputs != null)
			{
				throw new BlockInvalidException("Invalid coinbase TX (inputs != null");
			}
			if(tx.outputs.size() != 1)
			{
				throw new BlockInvalidException("Invalid coinbase TX (too many outputs)");
			}
			TransactionOutput txOutput = tx.outputs.iterator().next();
			if(txOutput.getValue() != Constants.coinbaseAmount)
			{
				throw new BlockInvalidException("Invalid coinbase TX (amount)");
			}
		}
		else
		{
			throw new BlockInvalidException("Frist TX is not a coinbase TX");
		}
		//check other values TODO
		
	}
	
	public String getPreviousBlockHash() {
		return previousHashChain2;
	}
	
	@Override
	public String toString() {
		return "CB [hash=" + hash + ", previousHashChain2=" + previousHashChain2 + ", lastBlockChain1Hash="
				+ lastBlockChain1Hash + ", electionHash=" + electionHash + ", merkleRoot=" + merkleRoot + ", transactions="
						+ transactions + ", timeStamp=" + timeStamp + ", nonce=" + nonce + "]";
	}
	
//	@Override
//	public String toString() {
//		return "CB [hash=" + hash + ", previousHashChain2=" + previousHashChain2 + ", previousBlockChain1Hash="
////				+ voteQuestion.getHashOfPreviousBlock() + ", subsequentBlockChain1Hash=" + voteQuestion.getHashOfSubsequentBlock()
//				+ ", lastBlockChain1Hash=" + lastBlockChain1Hash + ", merkleRoot=" + merkleRoot + ", transactions="
//				+ transactions + ", timeStamp=" + timeStamp + ", nonce=" + nonce + "]";
//	}

//	public final String getPreviousHashChain2() {
//		return previousHashChain2;
//	}

//	public final String getPreviousBlockChain1Hash() {
//		return previousBlockChain1Hash;
//	}
//
//	public final String getSubsequentBlockChain1Hash() {
//		return subsequentBlockChain1Hash;
//	}


	public final String getLastBlockChain1Hash() {
		return lastBlockChain1Hash;
	}

	public String getElectionHash() {
		return electionHash;
	}

//	public VoteQuestion getVoteQuestion() {
//		return voteQuestion;
//	}

}
