package redactableBlockchain.data;

import java.io.Serializable;
import java.math.BigInteger;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import redactableBlockchain.data.transaction.PaymentTransaction;
import redactableBlockchain.data.transaction.Transaction;
import redactableBlockchain.data.transaction.TransactionOutput;
import redactableBlockchain.data.transaction.VotingTransaction;
import redactableBlockchain.exception.BlockInvalidException;
import redactableBlockchain.exception.VerificationException;
import redactableBlockchain.util.Constants;
import redactableBlockchain.util.StringUtil;

public class StandardBlock extends Block implements Serializable{

	/**
	 * 
	 */
	private static final long serialVersionUID = -6803986682848737624L;
	public String previousHash;
	public String lastRedactionBlockHash; 

	private static final Logger logger = LoggerFactory.getLogger(StandardBlock.class);
	//Block Constructor.
	public StandardBlock(String previousHash, String lastRedactionBlockHash, Transaction coinbase) {
		super(BlockType.NormalBlock);
		this.previousHash = previousHash;
		this.lastRedactionBlockHash = lastRedactionBlockHash;
		this.timeStamp = new Date().getTime();
		this.transactions.add(coinbase);
		merkleRoot = StringUtil.getMerkleRoot(transactions);
//		this.hash = calculateHash(); //Making sure we do this after we set the other values.
	}

	//Calculate new hash based on blocks contents
	public String calculateHash() {
		String calculatedhash = StringUtil.applySha256( 
				previousHash +
				lastRedactionBlockHash +
				Long.toString(timeStamp) +
				Long.toString(nonce) + 
				merkleRoot
				);
		this.hash = calculatedhash;
		return calculatedhash;
	}

	//Add transactions to this block
	public boolean addTransaction(Transaction transaction) throws VerificationException {
		//process transaction and check if valid, unless block is genesis block then ignore.
		if(transaction == null) return false;		
		if((!"0000000000000000000000000000000000000000000000000000000000000000".equals(previousHash))) {
			if((transaction.processTransaction() != true)) {
				logger.warn("Transaction failed to process. Discarded.");
				return false;
			}
		}
		transactions.add(transaction);
		logger.info("Transaction Successfully added to Block");
		merkleRoot = StringUtil.getMerkleRoot(transactions);
		return true;
	}


	public String getPreviousBlockHash() {
		return previousHash;
	}

	public String getLastRedactionBlockHash() {
		return lastRedactionBlockHash;
	}

	public void validate() throws BlockInvalidException {
		//throws Exception ???
		calculateHash();
		if(getDiff().compareTo(getTargetDiff()) != -1)
		{
			System.out.println(calculateHash());
			System.out.println(getDiff());
			System.out.println(getTargetDiff());
			throw new BlockInvalidException("Difficulty too low...");
		}
		//TODO check block
		if(getTransactions().size() == 0)
		{
			throw new BlockInvalidException("No (coinbase) transaction in block");
		}
		Transaction claimedCoinBaseTX = getTransactions().iterator().next();
		if(claimedCoinBaseTX instanceof PaymentTransaction)
		{
			PaymentTransaction tx = (PaymentTransaction) claimedCoinBaseTX;
			if(tx.inputs != null)
			{
				throw new BlockInvalidException("Invalid coinbase TX (inputs != null");
			}
			if(tx.outputs.size() != 1)
			{
				throw new BlockInvalidException("Invalid coinbase TX (too many outputs)");
			}
			TransactionOutput txOutput = tx.outputs.iterator().next();
			if(txOutput.getValue() != Constants.coinbaseAmount)
			{
				throw new BlockInvalidException("Invalid coinbase TX (amount)");
			}
		}
		else
		{
			throw new BlockInvalidException("Frist TX is not a coinbase TX");
		}
		//TODO check included TX
		Iterator<Transaction> it = getTransactions().iterator();
		int votingTXcount = 0;
		while(it.hasNext())
		{
			Transaction tx = it.next();
			try {
				tx.validate();
				if(tx instanceof VotingTransaction)
					votingTXcount++;
				
			} catch (VerificationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				throw new BlockInvalidException(e);
			}
			if(votingTXcount >1)
				throw new BlockInvalidException("Too many voting TX in block");
		}
	}

	@Override
	public String toString() {
		return "SB [hash=" + hash + ", previousHash=" + previousHash + ", lastRedactionBlockHash="
				+ lastRedactionBlockHash + ", merkleRoot=" + merkleRoot + ", transactions=" + transactions
				+ ", timeStamp=" + timeStamp + ", nonce=" + nonce + "]";
	}
	
}
