package redactableBlockchain.data.transaction;
import java.io.Serializable;
import java.security.*;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.UUID;

//import javax.xml.bind.ValidationException;

import org.bouncycastle.util.encoders.Base64;

import redactableBlockchain.Wallet;
import redactableBlockchain.data.TransactionType;
import redactableBlockchain.exception.VerificationException;
import redactableBlockchain.util.Constants;
import redactableBlockchain.util.StringUtil;
import redactableBlockchain.util.Utils;


public class ElectionTransaction extends Transaction implements Serializable{
	
	private VoteQuestion voteQuestion;	

	public ElectionTransaction(PublicKey from, VoteQuestion voteQuestion) throws NoSuchAlgorithmException, InvalidKeySpecException {
		
	super(TransactionType.ELECTION);
		this.voteQuestion = voteQuestion;
		this.sender = Utils.PublicKeyToBase64(from);
	}
	
	public boolean processTransaction() throws VerificationException {
		
		if(verifySignature() == false) {
			System.out.println("#Transaction Signature failed to verify");
			return false;
		}		
		return true;
	}
	
	public long getInputsValue() {
		int total = 0;
//		for(TransactionInput i : inputs) {
//			if(i.UTXO == null) continue; //if Transaction can't be found skip it, This behavior may not be optimal.
//			total += i.UTXO.value;
//		}
		return total;
	}
	
	public ArrayList<TransactionInput> getInputs() {
		return inputs;
	}
	
	public ArrayList<TransactionOutput> getOutputs() {
		return outputs;
	}
	
//	public void generateSignature(Wallet wallet) throws NoSuchAlgorithmException, InvalidKeySpecException {
//		this.sender = Utils.PublicKeyToBase64(wallet.getPublicKey());
////		String data = sender + recipient + Float.toString(value);
//		String data = this.sender + this.data + this.voteQuestion.hashOfBlockToReplace;//TODO
//		signature = wallet.sign(data);
////		signature = StringUtil.applyECDSASig(wallet.privateKey,data);		
//	}
	
	public boolean verifySignature() throws VerificationException {
		try {
		//String data = this.sender + this.data + this.voteQuestion.hashOfBlockToReplace;//TODO
			String data = calculateHash();
		PublicKey publicKeySender;
		
			publicKeySender = Utils.PublicKeyFromBase64(sender);
			return StringUtil.verifyECDSASig(publicKeySender, data, signature);
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
			throw new VerificationException(e);
		} catch (InvalidKeySpecException e) {
			e.printStackTrace();
			throw new VerificationException(e);
		}
		
	}
	
	public long getOutputsValue() {
		int total = 0;
//		for(TransactionOutput o : outputs) {
//			total += o.value;
//		}
		return total;
	}
	
	public boolean isCoinbase() {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public String calculateHash() {
		return StringUtil.applySha256(
				data + 
				sender +
				voteQuestion.calculateHash()
				);
	}


	@Override
	protected void specificChecks() throws VerificationException {
		// TODO Auto-generated method stub
		
	}

	public VoteQuestion getVoteQuestion() {
		return voteQuestion;
	}

	@Override
	public String toString() {
		return "ElectionTransaction{" +
				"voteQuestion=" + voteQuestion +
				'}';
	}
}
