package redactableBlockchain.data.transaction;

import java.io.Serializable;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.UUID;

import redactableBlockchain.data.TransactionType;
import redactableBlockchain.exception.VerificationException;
import redactableBlockchain.util.StringUtil;
import redactableBlockchain.util.Utils;

public abstract class Transaction implements Serializable{

	private static final long serialVersionUID = -4392388885606161454L;
	protected TransactionType transactionType;
	public String data;
	public String sender; //Senders address/public key.
	//	public String recipient; //Recipients address/public key.
	public byte[] signature; //This is to prevent anybody else from spending funds in our wallet.
	//	public int value; //Contains the amount we wish to send to the recipient.
	public ArrayList<TransactionInput> inputs = new ArrayList<TransactionInput>();
	public ArrayList<TransactionOutput> outputs = new ArrayList<TransactionOutput>();
	public UUID uuid;

	public Transaction(TransactionType transactionType)
	{
		this.transactionType = transactionType;
		this.uuid = UUID.randomUUID();
	}

	public void sign(PrivateKey privateKey)
	{
		String hash = calculateHash();
		this.signature = StringUtil.applyECDSASig(privateKey, hash);
	}


	public abstract String calculateHash();
	//public abstract void validateSignature() throws VerificationException;

	public boolean verifySignature() throws VerificationException {
		try {
			String data = calculateHash();
			PublicKey publicKeySender = Utils.PublicKeyFromBase64(sender);
			return StringUtil.verifyECDSASig(publicKeySender, data, signature);
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
			throw new VerificationException(e);
		} catch (InvalidKeySpecException e) {
			e.printStackTrace();
			throw new VerificationException(e);
		}
	}

	protected abstract void specificChecks() throws VerificationException;

	public void validate() throws VerificationException
	{
		calculateHash();
		verifySignature();
		specificChecks();
	}

	public abstract boolean processTransaction() throws VerificationException;
	public abstract long getInputsValue();
	public abstract long getOutputsValue();

	public ArrayList<TransactionInput> getInputs() {
		return inputs;
	}

	public ArrayList<TransactionOutput> getOutputs() {
		return outputs;
	}

}
