package redactableBlockchain.data.transaction;

import java.io.Serializable;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;

import redactableBlockchain.util.StringUtil;
import redactableBlockchain.util.Utils;

public class TransactionOutput implements Serializable{
	/**
	 * 
	 */
	private static final long serialVersionUID = -5080188083041704358L;
	private String recipient; //hash or address of new owner
	private long value; //the amount of coins they own
	private UnlockType unlockType;
	
	private String hash = null;
	//Constructor
//	public TransactionOutput(PublicKey reciepient, int value, String parentTransactionHash) throws NoSuchAlgorithmException, InvalidKeySpecException {
//		this.recipient = Utils.PublicKeyToBase64(reciepient);
//		this.value = value;
//		this.parentTransactionHash = parentTransactionHash;
//		this.id = StringUtil.applySha256(this.recipient+Float.toString(value)+parentTransactionHash);
//	}
	public TransactionOutput(String recipientHash, long value) {
		this.recipient = recipientHash;
		this.value = value;
		this.unlockType = UnlockType.PUBLIC_KEY_HASH;
	}
	
	public TransactionOutput(PublicKey recipient, long value) throws NoSuchAlgorithmException, InvalidKeySpecException {
		this.recipient = StringUtil.applySha256(Utils.PublicKeyToBase64(recipient));
		this.value = value;
		this.unlockType = UnlockType.PUBLIC_KEY_HASH;
	}
	
	public TransactionOutput(String recipientHash, long value, UnlockType unlockType) {
		this.recipient = recipientHash;
		this.value = value;
		this.unlockType = unlockType;
	}
	
	public TransactionOutput(PublicKey recipient, long value, UnlockType unlockType) throws NoSuchAlgorithmException, InvalidKeySpecException {
		this.recipient = StringUtil.applySha256(Utils.PublicKeyToBase64(recipient));
		this.value = value;
		this.unlockType = unlockType;
	}
	
	public String calculateHash() {
		if(hash == null)
			hash = StringUtil.applySha256(
					recipient +
					value +
					unlockType
					);
		return hash;
	}
	
	//Check if coin belongs to you
	//does not actually verify/prove it
	public boolean isMine(PublicKey publicKey) throws NoSuchAlgorithmException, InvalidKeySpecException {
		switch(unlockType)
		{
		case PUBLIC_KEY:
			return (recipient.equalsIgnoreCase(Utils.PublicKeyToBase64(publicKey)));
		case PUBLIC_KEY_HASH:
			return (recipient.equalsIgnoreCase(Utils.getPublicKeyHash(publicKey)));
		default:
			return false;
		}
	}
	
	public long getValue() {
		return value;
	}

	/**
	 * @return the recipient
	 */
	public final String getRecipient() {
		return recipient;
	}

	/**
	 * @return the unlockType
	 */
	public final UnlockType getUnlockType() {
		return unlockType;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "TxOutput [recip=" + recipient + ", value=" + value + ", unlockType=" + unlockType + "]";
	}
	
	
}
