package redactableBlockchain.util;

import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;
import java.util.UUID;

import redactableBlockchain.Wallet;
import redactableBlockchain.data.CorrectionBlock;
import redactableBlockchain.data.StandardBlock;
import redactableBlockchain.data.transaction.PaymentTransaction;
import redactableBlockchain.data.transaction.TransactionOutput;

public class Constants {

	public static final long minimumTransactionAmount = 1;
	public static final long coinbaseAmount = 100;

	public final String genesisHash = "";
	public final long genesisNonce = 0;
	private static Wallet coinBaseWallet = null;
	private static Wallet aliceWallet = null;
	//public static long difficulty = 16*16*16*16*16*16*16*16l;
	public static long difficulty = 16;//16
	private static StandardBlock genesisBlock = null;
	private static CorrectionBlock genesisRedactionBlock = null;
	private static HashMap<String, Wallet> map = new HashMap<>();
	public static final int votingDuration = 3;//should be an odd number so we always get more or less than 50%

	public static StandardBlock getGenesisBlock()
	{
		try {
			if(genesisBlock == null)
			{
				PaymentTransaction genesisTransaction = createGenesisCoinBasesTransaction(getAliceWallet().getPublicKey());
				StandardBlock genesis = new StandardBlock("0000000000000000000000000000000000000000000000000000000000000000", "0000000000000000000000000000000000000000000000000000000000000000", genesisTransaction);
				genesis.timeStamp = 1526152683837l;

//							genesis.addTransaction(genesisTransaction);
//				genesis.setNonce(1417045l);//17368l;
				genesis.setNonce(2534);//17368l;
//				genesis.prepareBlockForMining();
				genesis.calculateHash();
				genesisBlock = genesis;
//				System.out.println("Genesis Hash:"+genesisBlock.hash);
				System.out.println("Genesis Hash:"+genesisBlock.calculateHash());
			}
		}catch(Exception e)
		{
			e.printStackTrace();
		}
		return genesisBlock;
	}

	public static CorrectionBlock getGenesisRedactionBlock() throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		if(genesisRedactionBlock == null)
		{
			PaymentTransaction genesisTransaction = createGenesisCoinBasesTransaction(getAliceWallet().getPublicKey());
			CorrectionBlock genesisRedactionBlock1 = new CorrectionBlock("0000000000000000000000000000000000000000000000000000000000000000", "0000000000000000000000000000000000000000000000000000000000000000", "0000000000000000000000000000000000000000000000000000000000000000", "0000000000000000000000000000000000000000000000000000000000000000", "0000000000000000000000000000000000000000000000000000000000000000", genesisTransaction);
			genesisRedactionBlock1.timeStamp = 1526152683837l;

			//			genesis.addTransaction(genesisTransaction);
//			genesisRedactionBlock1.prepareBlockForMining();
			genesisRedactionBlock1.setNonce(2476384473l);
			genesisRedactionBlock1.calculateHash();
			System.out.println("Genesis Hash:"+genesisRedactionBlock1.getHash());
			genesisRedactionBlock = genesisRedactionBlock1;
		}
		return genesisRedactionBlock;
	}

	public static Wallet getCoinbaseWallet()
	{
		if(coinBaseWallet == null)
		{
			coinBaseWallet = new Wallet("coinbase");
		}
		return coinBaseWallet;
	}

	public static Wallet getAliceWallet()
	{
		if(aliceWallet == null)
		{	
			aliceWallet = new Wallet("Alice");
		}
		return aliceWallet;
	}

	public static Wallet getRandomWallet()
	{
//		String[] seeds = {"Alice","Bob","1","2","3","4","5"};
		String[] seeds = {"1","2","3","4","5"};
		Random random = new Random();
		int index = random.nextInt(seeds.length);
		//return new Wallet(seeds[index]);
		return getWallet(seeds[index]);
	}

	public static Wallet getWallet(String id)
	{
		if(map.get(id) == null)
		{	
			map.put(id, new Wallet(id));
		}
		return map.get(id);
	}
	public static PaymentTransaction createGenesisCoinBasesTransaction(PublicKey receiver) throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		Wallet coinbase = getCoinbaseWallet();
		PaymentTransaction tx = createCoinBaseTransaction(receiver);
		tx.uuid = UUID.fromString("16f41d3c-c56b-4245-a596-cce1d55eb00c");
		//tx.generateSignature(coinbase.privateKey);	 //manually sign the genesis transaction
		byte[] sig = {48, 53, 2, 25, 0, -79, 52, 49, -124, -84, 112, -100, -88, 114, 108, 65, -51, -109, -110, 63, 23, 107, 16, -76, -39, -102, 40, -9, 34, 2, 24, 110, -21, 123, -74, 19, -44, -106, -35, 114, -125, 51, -61, 77, 28, -47, -54, 119, -97, 88, -52, -120, 85, 92, -53};
		tx.signature = new byte[sig.length];
		System.arraycopy(sig, 0, tx.signature, 0, sig.length);
		tx.setSignature(sig);
		System.out.println("tx hash:"+tx.calculateHash());
		return tx;
	}

	public static PaymentTransaction createCoinBaseTransaction(PublicKey receiver) throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		Wallet coinbase = getCoinbaseWallet();
		ArrayList<TransactionOutput> outputs = new ArrayList<>();
		TransactionOutput output = new TransactionOutput(receiver, coinbaseAmount);
		outputs.add(output);
		//create coinbase transaction, which sends 100 [coinbaseAmount] coins to receiver: 
		PaymentTransaction genesisTransaction = new PaymentTransaction(coinbase.getPublicKeyAsString(), null, outputs );
//		PaymentTransaction genesisTransaction = new PaymentTransaction(coinbase.getPublicKey(), receiver, coinbaseAmount, null);
//		genesisTransaction.transactionId = "0"; //manually set the transaction id
		//genesisTransaction.outputs.add(new TransactionOutput(genesisTransaction.recipient, genesisTransaction.value, "0000000000000000000000000000000000000000000000000000000000000000")); //manually add the Transactions Output
		genesisTransaction.sign(coinbase.privateKey);	 //manually sign the coinbase transaction	
		return genesisTransaction;	
	}

	public static int votingMinVotes() {
		return ((int)(Constants.votingDuration/2)+1);
	}
}
