package redactableBlockchain.util;

import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;

import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;

import org.apache.commons.codec.binary.Base64;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import redactableBlockchain.data.Block;
import redactableBlockchain.data.IBlock;
import redactableBlockchain.data.CorrectionBlock;
import redactableBlockchain.data.StandardBlock;
import redactableBlockchain.data.transaction.ElectionTransaction;
import redactableBlockchain.data.transaction.PaymentTransaction;
import redactableBlockchain.data.transaction.Transaction;
import redactableBlockchain.data.transaction.VotingTransaction;

public class Utils {

	static Gson gson = null;

	static
	{
		RuntimeTypeAdapterFactory<Transaction> adapter = 
				RuntimeTypeAdapterFactory
				.of(Transaction.class)
				.registerSubtype(ElectionTransaction.class)
				.registerSubtype(PaymentTransaction.class)
				.registerSubtype(VotingTransaction.class);

		adapter.of(Block.class)
		.registerSubtype(StandardBlock.class)
		.registerSubtype(CorrectionBlock.class);


		gson = new GsonBuilder().registerTypeAdapterFactory(adapter).create();
	}


	public static String ObjectToString(Object o)
	{
		return gson.toJson(o);
	}

	public static Object ObjectfromString(String json, Class classOfT)
	{
		return gson.fromJson(json, classOfT);
	}

	public static String PublicKeyToBase64(PublicKey pub) throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		KeyFactory keyFactory = KeyFactory.getInstance("ECDSA");
		X509EncodedKeySpec keySpec = keyFactory.getKeySpec(pub, X509EncodedKeySpec.class);

		Base64 codec = new Base64();
		byte[] encoded = codec.encode(keySpec.getEncoded());
		return new String(encoded);
	}

	public static PublicKey PublicKeyFromBase64(String sender) throws NoSuchAlgorithmException, InvalidKeySpecException {
		Base64 codec = new Base64();
		byte[] encodedKey = codec.decode(sender);
		X509EncodedKeySpec pubSpec=new X509EncodedKeySpec(encodedKey);
		KeyFactory keyFactory = KeyFactory.getInstance("ECDSA");
		return keyFactory.generatePublic(pubSpec);

	}
	
	public static String getPublicKeyHash(PublicKey pub) throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		String val = PublicKeyToBase64(pub);
		return getPublicKeyHashFromBase64(val);
	}
	
	public static String getPublicKeyHashFromBase64(String val) throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		return StringUtil.applySha256(val);
	}
}
