package redactableBlockchain.visualization;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.UUID;

import guru.nidi.graphviz.attribute.Color;
import guru.nidi.graphviz.attribute.Style;
import guru.nidi.graphviz.engine.Format;
import guru.nidi.graphviz.engine.Graphviz;
import guru.nidi.graphviz.model.MutableGraph;
import guru.nidi.graphviz.parse.Parser;
import redactableBlockchain.data.*;
import redactableBlockchain.data.transaction.ElectionTransaction;
import redactableBlockchain.data.transaction.Transaction;
import redactableBlockchain.data.transaction.VoteQuestion;
//import redactableBlockchain.data.FakeBlock;

//import redactableBlockchain.data.ModifyBlock;
//import redactableBlockchain.data.NormalBlock;

public class CreatePicture {

    static int pictureNumber = 0;
    final static String prefix = "digraph G {\r\n" +
            "compound=true;\r\n" +
            "concentrate=true;\r\n" +
            "    graph [fontsize=10 fontname=\"Verdana\"];\r\n" +
            "    node [shape=record fontsize=10 fontname=\"Verdana\"];\r\n" +
            "\r\n" +
            "";
    final static String postfix = "}";

    public static void main(String[] args) throws IOException {
        createPNG("graph {\r\n" +
                "    { rank=same; white}\r\n" +
                "    { rank=same; cyan; yellow; pink}\r\n" +
                "    { rank=same; red; green; blue}\r\n" +
                "    { rank=same; black}\r\n" +
                "\r\n" +
                "    white -- cyan -- blue\r\n" +
                "    white -- yellow -- green\r\n" +
                "    white -- pink -- red\r\n" +
                "\r\n" +
                "    cyan -- green -- black\r\n" +
                "    yellow -- red -- black\r\n" +
                "    pink -- blue -- black\r\n" +
                "}","test");

        createPNG("digraph G {\r\n" +
                "\r\n" +
                "    graph [fontsize=10 fontname=\"Verdana\"];\r\n" +
                "    node [shape=record fontsize=10 fontname=\"Verdana\"];\r\n" +
                "\r\n" +
                "    subgraph cluster_0 {\r\n" +
                "        node [style=filled];\r\n" +
                "        \"Item 1\" \"Item 2\";\r\n" +
                "        label = \"Container A\";\r\n" +
                "        color=blue;\r\n" +
                "    }\r\n" +
                "\r\n" +
                "    subgraph cluster_1 {\r\n" +
                "        node [style=filled];\r\n" +
                "        \"Item 3\" \"Item 4\";\r\n" +
                "        label = \"Container B\";\r\n" +
                "        color=blue;\r\n" +
                "    }\r\n" +
                "\r\n" +
                "    subgraph cluster_2 {\r\n" +
                "        node [style=filled];\r\n" +
                "        \"Item 5\" \"Item 6\";\r\n" +
                "        label = \"Container C\";\r\n" +
                "        color=blue;\r\n" +
                "    }\r\n" +
                "\r\n" +
                "    // Renders fine\r\n" +
                "    \"Item 1\" -> \"Item 2\";\r\n" +
                "    \"Item 2\" -> \"Item 3\";\r\n" +
                "\r\n" +
                "    // Both of these create new nodes\r\n" +
                "    cluster_1 -> cluster_2;\r\n" +
                "    \"Container A\" -> \"Container C\";\r\n" +
                "}","test2");

    }

    public static void createPNG(HashMap<String, IBlock> normalBlocks, HashMap<String, CorrectionBlock> modifyBlocks, String nodeID, Chain longestChain) throws Exception {
        String dot = createDot(normalBlocks, modifyBlocks, longestChain);
        System.out.println(dot);
        createPNG(dot, nodeID);
    }

    static HashMap<String,ElectionTransaction> electionMap = new HashMap<>();
    public static String createDot(HashMap<String, IBlock> normalBlocks, HashMap<String, CorrectionBlock> modifyBlocks, Chain longestChain) throws Exception {

        //search for elections
        electionMap.clear();
        for(IBlock b : normalBlocks.values())
        {
            for(Transaction tx : b.getTransactions())
            {
                if(tx instanceof ElectionTransaction)
                {
                    electionMap.put(tx.calculateHash(), (ElectionTransaction) tx);
                }
            }
        }
        StringBuilder sb = new StringBuilder();
        sb.append(prefix);

        IBlock lastBlock = longestChain.getLastBlock();
        HashSet<String> blocks = new HashSet<>();
        HashSet<String> links = new HashSet<>();
        while(!lastBlock.getPreviousBlockHash().equalsIgnoreCase("0000000000000000000000000000000000000000000000000000000000000000"))
        {
            blocks.add(lastBlock.calculateHash());
            links.add(lastBlock.getPreviousBlockHash()+lastBlock.calculateHash());
            lastBlock = longestChain.get(lastBlock.getPreviousBlockHash());
        }
        //add all blocks(nodes
        int i=0;
        for(IBlock block_ : normalBlocks.values())
        {
            if(block_ instanceof StandardBlock)
            {
                StandardBlock block = (StandardBlock) block_;
                String hash = getShortHash(block.calculateHash());
                String last2 = getShortHash(block.getLastRedactionBlockHash());
                String prevBlock = getShortHash(block.getPreviousBlockHash());

                //struct4 [label="{<hash>hash |{<prev> prev|...|<e> lastM}}"];
                if(blocks.contains(block.getHash()))
                {
                    sb.append("H"+block.getHash() + " [label=\"{<hash> "+hash + " |{<prev> " + prevBlock + "|...|<last2> " + last2 + "}}\",color=green];\n");
                }
                else
                {
                    sb.append("H"+block.getHash() + " [label=\"{<hash> "+hash + " |{<prev> " + prevBlock + "|...|<last2> " + last2 + "}}\"];\n");
                }

                if(!block.getPreviousBlockHash().equalsIgnoreCase("0000000000000000000000000000000000000000000000000000000000000000") && !block.getLastRedactionBlockHash().equalsIgnoreCase("0000000000000000000000000000000000000000000000000000000000000000"))
                {
                    if(links.contains(block.getPreviousBlockHash()+block.getHash()))
                    {
                        sb.append("H"+block.getHash()+":prev -> H" + block.getPreviousBlockHash()+":hash [color=green];\n");
                    }
                    else
                    {
                        sb.append("H"+block.getHash()+":prev -> H" + block.getPreviousBlockHash()+":hash;\n");
                    }
                    sb.append("H"+block.getHash()+":last2 -> H" + block.getLastRedactionBlockHash()+":hash;\n");
                }

//				sb.append("subgraph cluster").append(block.getHash()).append("{\n");
//				sb.append("label = \"").append(hash+"\";\n");
//				String uuid = UUID.randomUUID().toString();
//				sb.append("\"").append("P_").append(prevBlock).append("\" \"").append("LM_").append(uuid).append("\" [label=\"LM_").append(lastM).append("\"];\n");
//				sb.append("\"").append("H_").append(hash).append("\";\n");
//				sb.append("}\n");
//				i++;
//
//				//add edge to previous block
//				sb.append("\"P_").append(prevBlock).append("\" -> \"H_").append(prevBlock).append("\" [lhead=cluster").append(block.getPreviousBlockHash()).append("];\n");
//
//
//				//add edge to last modification block
//				sb.append("\"LM_").append(uuid).append("\" -> \"H_").append(lastM).append("\" [lhead=cluster").append(block.getLastRedactionBlockHash()).append("];\n");
            }
            else if(block_ instanceof VirtualBlock)
            {
                VirtualBlock block = (VirtualBlock) block_;
                String hash = getShortHash(block.getHash());
                String last2 = getShortHash(block.getLastRedactionBlockHash());
                String prevBlock = getShortHash(block.getPreviousBlockHash());

                //struct4 [label="{<hash>hash |{<prev> prev|...|<e> lastM}}"];
                if(blocks.contains(block.getHash()))
                {
                    sb.append("H"+block.getHash() + " [label=\"{<hash> "+hash + " |{<prev> " + prevBlock + "|...|<last2> " + last2 + "}}\",color=orange];\n");
                }
                else
                {
                    sb.append("H"+block.getHash() + " [label=\"{<hash> "+hash + " |{<prev> " + prevBlock + "|...|<last2> " + last2 + "}}\",color=grey];\n");
                }

                if(!block.getPreviousBlockHash().equalsIgnoreCase("0000000000000000000000000000000000000000000000000000000000000000") && !block.getLastRedactionBlockHash().equalsIgnoreCase("0000000000000000000000000000000000000000000000000000000000000000"))
                {
                    if(links.contains(block.getPreviousBlockHash()+block.getHash()))
                    {
                        sb.append("H"+block.getHash()+":prev -> H" + block.getPreviousBlockHash()+":hash [color=orange];\n");
                    }
                    else
                    {
                        sb.append("H"+block.getHash()+":prev -> H" + block.getPreviousBlockHash()+":hash;\n");
                    }
                    sb.append("H"+block.getHash()+":last2 -> H" + block.getLastRedactionBlockHash()+":hash [color=magenta,dir=both];\n");
                }
            }
            else
            {
                System.err.println("Should never occour?");
//                assertTrue(false);
                throw new Exception("TODO");
            }
        }
        for(CorrectionBlock block : modifyBlocks.values())
        {
            //int i = 0;
            addModifyBlock(block, sb, i);

        }
        //add all edges


        sb.append(postfix);
        return sb.toString();
    }

    // struct3 [label="{<hash> hash|{<prev2> prev 2|<other> ...|<last1> last chain 1}| {<prev1> prev chain 1| <sub1>sub chain 1}}",color=red];

    //struct4 [label="{<hash>hash |{<c> prev|...|<e> lastM}}"];

    private static void addModifyBlock(CorrectionBlock block, StringBuilder sb, int i) {
        String hash = getShortHash(block.getHash());
//		String prevBlock = getShortHash(block.getPreviousBlockHash());

        ElectionTransaction eTX = electionMap.get(block.getElectionHash());
        VoteQuestion voteQuestion = eTX.getVoteQuestion();

        String prevReplace = getShortHash(voteQuestion.getHashOfPreviousBlock());
        String subsReplace = getShortHash(voteQuestion.getHashOfSubsequentBlock());

        String lastChain1 = getShortHash(block.getLastBlockChain1Hash());
        //TODO get election
        String prevChain2 = getShortHash(block.getPreviousBlockHash());
        String electionHash = getShortHash(block.getElectionHash());

        //struct3 [label="{<hash> hash|{<prev2> prev 2|<other> ...|<last1> last chain 1}| {<prev1> prev chain 1| <sub1>sub chain 1}}",color=red];
        sb.append("H" + block.getHash() + " [label=\"{<hash> " + hash + "|{<prev2> " + prevChain2 + "|<other> ...|<last1> " + lastChain1 + "}| {<prev1> " + prevReplace + " | <subs1> " + subsReplace + "}}\",color=red];\n");

//		sb.append("subgraph cluster").append(block.getHash()).append("{\n");
//		sb.append("label = \"").append(hash+"\";\n");
//		String uuid = UUID.randomUUID().toString();
//		sb.append("\"").append("P_").append(prevChain2).append("\" ");//TODO not unique
//		sb.append("\"").append("P1_").append(prevReplace).append("\" ");
//		sb.append("\"").append("S1_").append(subsReplace).append("\" ");
//		sb.append("\"").append("L1_").append(lastChain1).append("\" ");
////				+ ""
////				+ "\"").append("LM_").append(uuid).append("\" [label=\"LM_").append(lastM).append("\"] ");
//		sb.append("\"").append("H_").append(hash).append("\";\n");
//		sb.append("color=blue\n");
//		sb.append("}\n");
//		i++;

        //add edges

        sb.append("H" + block.getHash()+":last1 -> H" + block.getLastBlockChain1Hash()+":hash [color=blueviolet];\n");
        sb.append("H" + block.getHash()+":prev2 -> H" + block.getPreviousBlockHash()+":hash;\n");
        sb.append("H" + block.getHash()+":prev1 -> H" + voteQuestion.getHashOfPreviousBlock()+":hash [color=coral];\n");
        sb.append("H" + block.getHash()+":subs1 -> H" + voteQuestion.getHashOfSubsequentBlock()+":hash [color=dodgerblue2];\n");

//
//		sb.append("\"P_").append(prevChain2).append("\" -> \"H_").append(prevChain2).append("\" [lhead=cluster").append(block.getPreviousHashChain2()).append("];\n");//TODO not unique
//
//		sb.append("\"L1_").append(lastChain1).append("\" -> \"H_").append(lastChain1).append("\" [lhead=cluster").append(block.getLastBlockChain1Hash()).append("];\n");
//
//		sb.append("\"P1_").append(prevReplace).append("\" -> \"H_").append(prevReplace).append("\" [lhead=cluster").append(block.getVoteQuestion().getHashOfPreviousBlock()).append("];\n");
//		sb.append("\"S1_").append(subsReplace).append("\" -> \"H_").append(subsReplace).append("\" [lhead=cluster").append(block.getVoteQuestion().getHashOfSubsequentBlock()).append("];\n");

    }

    public static void createPNG(String dot, String nodeID)
    {
        try {
            MutableGraph g = Parser.read(dot);
            Graphviz.fromGraph(g).render(Format.PNG).toFile(new File("example/"+nodeID+"_"+String.format("%04d", pictureNumber)+".png"));
            pictureNumber++;
            //			g.graphAttrs()
            //			.add(Color.WHITE.gradient(Color.rgb("888888")).background().angle(90))
            //			.nodeAttrs().add(Color.WHITE.fill())
            //			.nodes().forEach(node ->
            //			node.add(
            //					Color.named(node.name().toString()),
            //					Style.lineWidth(4).and(Style.FILLED)));
            //			Graphviz.fromGraph(g).width(700).render(Format.PNG).toFile(new File("example/"+nodeID+"_"+String.format("%04d", pictureNumber)+"-2.png"));

        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    private static String getShortHash(String hash)
    {
        return hash.substring(0, 5)+"..."+hash.substring(hash.length()-5);
    }

}