/*
 * This Kotlin source file was generated by the Gradle 'init' task.
 */
package at.tugraz.iaik.p2p.pol

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.pol.msg.RequestMessage
import at.tugraz.iaik.p2p.pol.msg.ResponseMessage
import at.tugraz.iaik.p2p.skad.config.Config
import at.tugraz.iaik.p2p.skad.msg.ExtendedMessage
import at.tugraz.iaik.p2p.skad.node.LocalNode
import at.tugraz.iaik.p2p.skad.node.Node
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.asCoroutineDispatcher
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import mu.KotlinLogging
import redactableBlockchain.IBlockChain
import redactableBlockchain.IBlockChainListener
import redactableBlockchain.KotlinUtils
import redactableBlockchain.UnoptimizedRedactableBlockchain.blockchain
import redactableBlockchain.data.CorrectionBlock
import redactableBlockchain.data.IBlock
import redactableBlockchain.data.StandardBlock
import redactableBlockchain.exception.VerificationException
import redactableBlockchain.network.data.*
import redactableBlockchain.util.Utils
import java.security.NoSuchAlgorithmException
import java.security.spec.InvalidKeySpecException
import java.util.concurrent.ForkJoinPool

//P2P Node capable of issuing decrypt requests and handling incoming policy evaluaton requests
class NetworkNode(private val config: Config, private val blockchain: IBlockChain) : IBlockChainListener {


    //kademlia node responsible for p2p communication with listener for custom messages
    private val node = LocalNode(config, incomingMsgListener = ::listen, actorCtx = KotlinUtils.actorCtx1, reactorCtx = KotlinUtils.reactorCtx1)//actorCtx = ForkJoinPool().asCoroutineDispatcher(),reactorCtx = ForkJoinPool().asCoroutineDispatcher())
    //    val logger = KotlinLogging.logger("P ${node.id}")
    val logger = KotlinLogging.logger("NetworkNode")
    val identifier = node.id

    init {
        blockchain.addListener(this)
//        this.blockchain = blockchain
    }

    override fun notifyMissingBlock(blockHash: String?) {
        //ask for missing block
        val getBlock = GetBlock(blockHash)
        logger.debug { "Requesting getBlock:$blockHash" }
        val tmp = Utils.ObjectToString(getBlock)
//        runBlocking {
        broadcastMessage(tmp)
//        }
    }

    override fun notifyAddedBlock(block: StandardBlock?) {
        //broadcast to all
        logger.debug("Broadcasting added block... ${block?.hash}")
        val tmp = Utils.ObjectToString(NormalBlockData(block)) + "\r\n"
//        runBlocking {
        broadcastMessage(tmp)
//        }
    }

    override fun notifyAddedRedactionBlock(block: CorrectionBlock?) {
        //do nothing?
        logger.debug("${this.identifier.toString()} Broadcasting added correction block... ${block?.hash}")
        val tmp = Utils.ObjectToString(ModifyBlockData(block)) + "\r\n"
//        runBlocking {
        broadcastMessage(tmp)
//        }
    }

    override fun notifyNewLongestChain(block: IBlock?) {
        //do nothing
    }

    override fun notifyNewBlockToMine(block: IBlock?) {
        //do nothing
    }

    //
    suspend fun bootstrap(node: NetworkNode) = this.node.bootstrap(node.node)

    fun broadcastMessage(message: String) {
        var nodes = node.getKnownNodes(sorted = false);
        for (nodeinfo in nodes) {
//            val requestMessage = RequestMessage(
//                    CommonCrypto.RND.nextLong(),
//                    node,
//                    message
//            )
//            logger.debug(this.node.id.toString() + " sending message to " + nodeinfo.node)
//            node.actor.send(requestMessage, nodeinfo.node)
            GlobalScope.launch { sendMessage(message, nodeinfo.node) }

        }
    }

    suspend fun sendMessage(message: String, otherNode: Node) {
        val requestMessage = RequestMessage(
                CommonCrypto.RND.nextLong(),
                node,
                message
        )
        logger.debug(this.node.id.toString() + " sending message to " + otherNode + " (" + message.length + " Bytes): " + message)
        node.actor.send(requestMessage, otherNode)
    }
//}

    private suspend fun listen(msg: ExtendedMessage): ExtendedMessage {
//        println(this.node.id.toString() + " received message "+ msg)

        val job = GlobalScope.launch {
            if (msg is RequestMessage) {
                //println(this.node.id.toString() + " received: " + msg.msg)
//                println("${this@NetworkNode.node.id.toString()} received: " + msg.msg)
                handleMessage(msg)
            } else {
                //println("ELSE " + msg.type)
                TODO("ELSE" + msg.type)
            }
        }
        return ResponseMessage(msg.seqId, this.node, "")
    }

    private fun handleMessage(requestMessage: RequestMessage) {
        val msg = requestMessage.msg
//        val a = msg.split("]".toRegex(), 2)[1]
//        println("Parsing:" + msg.split("]".toRegex(), 2)[1])
        val basedata = Utils.ObjectfromString(msg, BaseData::class.java) as BaseData
        when (basedata.type) {
            Type.NormalBLOCK -> addBlock(basedata)
            Type.ModifyBLOCK -> addModifyBlock(basedata)
//            Type.Heartbeat -> result = sendUpdates(basedata);
//            Type.TRANSACTION -> result = addTransaction(basedata);
            Type.GetBlock -> sendBlock(requestMessage.origin, basedata)
            else -> {
                println("Not supported:" + basedata.type)
            }
        }
    }

    private fun sendBlock(origin: Node, basedata: BaseData) {
//        val b = Utils.ObjectfromString(basedata.data, GetBlock::class.java) as GetBlock
        val block = blockchain.getBlock(basedata.data);
        if (block is StandardBlock)
            runBlocking { sendMessage(Utils.ObjectToString(NormalBlockData(block)), origin) }
        else if (block is CorrectionBlock)
            runBlocking { sendMessage(Utils.ObjectToString(ModifyBlockData(block)), origin) }
        else {

//            TODO("TODO")
        }

    }

    private fun addModifyBlock(basedata: BaseData) {
        val b = Utils.ObjectfromString(basedata.data, CorrectionBlock::class.java) as CorrectionBlock

        try {
            logger.debug { "${this.node.id.toString()} adding correction block ${b.hash}" }
            blockchain.addModifyBlock(b)
        } catch (e: NoSuchAlgorithmException) {
            e.printStackTrace()
        } catch (e: InvalidKeySpecException) {
            e.printStackTrace()
        } catch (e: VerificationException) {
            e.printStackTrace()
        }
    }

    private fun addBlock(basedata: BaseData) {
        val b = Utils.ObjectfromString(basedata.data, StandardBlock::class.java) as StandardBlock

        try {
            //add block to blockchain
            logger.debug { "${this.node.id.toString()} adding block ${b.hash}" }
            blockchain.addBlock(b)
        } catch (e: NoSuchAlgorithmException) {
            e.printStackTrace()
        } catch (e: InvalidKeySpecException) {
            e.printStackTrace()
        } catch (e: VerificationException) {
            e.printStackTrace()
        }
    }

    public suspend fun stop() {
        this.node.close();
    }
}