package at.tugraz.iaik.p2p.pol.msg

import at.tugraz.iaik.p2p.commons.readByteArray
import at.tugraz.iaik.p2p.commons.writeByteArray
import at.tugraz.iaik.p2p.commons.writeShort
import at.tugraz.iaik.p2p.pol.*
import at.tugraz.iaik.p2p.skad.id.ECID
import at.tugraz.iaik.p2p.skad.id.readIdentifier
import at.tugraz.iaik.p2p.skad.id.writeIdentifier
import at.tugraz.iaik.p2p.skad.msg.ExtendedMessage
import at.tugraz.iaik.p2p.skad.msg.ExtendedType
import at.tugraz.iaik.p2p.skad.msg.Message
import at.tugraz.iaik.p2p.skad.msg.MessageHeader
import at.tugraz.iaik.p2p.skad.node.Node
import java.io.DataInput
import java.io.DataOutput

/**
 * Custom messages for policy evaluation in top of kademlia base p2p network see message base classes for more information
 */
enum class Error {
    UNKNOWN_MSG, EVL_MISMATCH, INVALID_SHARE_SIG, INVALID_POLICY_SIG, UNAUTHORIZED
}

//Identifiers for custom message types; use some random long as base
private var offset = 645710602390000L
val T_ERR = ++offset
val T_PEV = ++offset
val T_PER = ++offset

abstract class BlockchainMessage(
        extendedType: ExtendedType,
        val acceptedType: ExtendedType,
        seqId: Long,
        origin: Node,
        deser: (header: MessageHeader, buf: DataInput) -> Message
) : ExtendedMessage(extendedType, seqId, origin, deser)

//class ErrorMessage(seqId: Long, origin: Node, val msg: String) :
//        ExtendedMessage(
//                T_ERR,
//                seqId,
//                origin,
//                deser = { rd, buffer ->
//                    ErrorMessage(
//                            rd.seqId,
//                            rd.origin,
//                            Error.values()[buffer.readShort().toUShort().toInt()],
//                            String(buffer.readByteArray())
//                    )
//                }) {
//    override fun serialize(buffer: DataOutput) {
//
//        //buffer.writeShort(code.ordinal.toUShort().toShort())
//        buffer.writeByteArray(msg.toByteArray())
//    }
//
//    override fun toString(): String {
//        return "ErrorMessage(seqId=$seqId, origin=$origin, code=$code, msg=\"$msg\")"
//    }
//}

class RequestMessage(seqId: Long, origin: Node, val msg: String) :
        BlockchainMessage(
                T_PEV, T_PER,
                seqId,
                origin,
                deser = { rd, buffer ->
//                    val bytes:ByteArray
//                    buffer.readFully(bytes)
                    RequestMessage(
                            rd.seqId,
                            rd.origin,
                            String(buffer.readByteArray())
                    )
                }) {
    override fun serialize(buffer: DataOutput) {
//        buffer.write(msg.toByteArray().size)
//        buffer.writeBytes(msg)
        buffer.writeByteArray(msg.toByteArray())
//        buffer.writeSignedPolicy(policy)
//        buffer.writeIdentifier(creator)
    }

}

class ResponseMessage(seqId: Long, origin: Node, val msg: String) :
        ExtendedMessage(
                T_PER,
                seqId,
                origin,
                deser = { rd, buffer ->
                    ResponseMessage(
                            rd.seqId,
                            rd.origin,
                            String(buffer.readByteArray())
                    )
                }) {
    override fun serialize(buffer: DataOutput) {
        buffer.writeByteArray(msg.toByteArray())
    }
}